package ru.dbotthepony.mc.otm.menu

import net.minecraft.network.FriendlyByteBuf
import net.minecraft.server.level.ServerPlayer
import net.minecraft.world.entity.player.Inventory
import net.minecraft.world.entity.player.Player
import net.minecraft.world.inventory.Slot
import net.minecraftforge.network.NetworkEvent
import net.minecraftforge.network.PacketDistributor
import ru.dbotthepony.mc.otm.OverdriveThatMatters
import ru.dbotthepony.mc.otm.block.entity.matter.MatterPanelBlockEntity
import ru.dbotthepony.mc.otm.capability.matter.MatterTask
import ru.dbotthepony.mc.otm.capability.matter.PatternState
import ru.dbotthepony.mc.otm.client.minecraft
import ru.dbotthepony.mc.otm.graph.matter.IMatterGraphListener
import ru.dbotthepony.mc.otm.graph.matter.MatterNetworkGraph
import ru.dbotthepony.mc.otm.network.*
import ru.dbotthepony.mc.otm.registry.MMenus
import java.util.*
import java.util.function.Consumer
import java.util.function.Supplier

class CancelTaskPacket(val id: UUID) : MatteryPacket {
	override fun write(buff: FriendlyByteBuf) {
		buff.writeUUID(id)
	}

	override fun play(context: Supplier<NetworkEvent.Context>) {
		context.packetHandled = true
		context.enqueueWork {
			val sender = context.sender!!
			(sender.containerMenu as? MatterPanelMenu)?.receiveTaskCancel(sender, id)
		}
	}

	companion object {
		fun read(buff: FriendlyByteBuf): CancelTaskPacket {
			return CancelTaskPacket(buff.readUUID())
		}
	}
}

class PatternsChangePacket(val isUpdate: Boolean, val patterns: Collection<PatternState>) : MatteryPacket {
	override fun write(buff: FriendlyByteBuf) {
		buff.writeBoolean(isUpdate)
		buff.writeInt(patterns.size)

		for (pattern in patterns) {
			pattern.write(buff)
		}
	}

	override fun play(context: Supplier<NetworkEvent.Context>) {
		context.packetHandled = true
		context.enqueueWork {
			val menu = minecraft.player?.containerMenu as? MatterPanelMenu ?: return@enqueueWork

			if (isUpdate) {
				menu.networkPatternsUpdated(patterns)
			} else {
				menu.networkPatternsRemoved(patterns)
			}
		}
	}

	companion object {
		fun read(buff: FriendlyByteBuf): PatternsChangePacket {
			val isUpdate = buff.readBoolean()
			val size = buff.readInt()

			return PatternsChangePacket(isUpdate, ArrayList<PatternState>(size).also {
				for (i in 0 until size) it.add(PatternState.read(buff) ?: throw NullPointerException("Unable to read PatternState from network at $i")) })
		}
	}
}

class TasksChangePacket(val isUpdate: Boolean, val tasks: Collection<MatterTask>) : MatteryPacket {
	override fun write(buff: FriendlyByteBuf) {
		buff.writeBoolean(isUpdate)
		buff.writeInt(tasks.size)

		for (task in tasks) {
			task.write(buff)
		}
	}

	override fun play(context: Supplier<NetworkEvent.Context>) {
		context.packetHandled = true
		context.enqueueWork {
			val menu = minecraft.player?.containerMenu as? MatterPanelMenu ?: return@enqueueWork

			if (isUpdate) {
				menu.networkTasksUpdated(tasks)
			} else {
				menu.networkTasksRemoved(tasks)
			}
		}
	}

	companion object {
		fun read(buff: FriendlyByteBuf): TasksChangePacket {
			val isUpdate = buff.readBoolean()
			val size = buff.readInt()

			return TasksChangePacket(isUpdate, ArrayList<MatterTask>(size).also {
				for (i in 0 until size) it.add(MatterTask.read(buff) ?: throw NullPointerException("Unable to read PatternState from network at $i")) })
		}
	}
}

class ReplicationRequestPacket(val state: PatternState, amount: Int) : MatteryPacket {
	val amount = amount.coerceAtLeast(1).coerceAtMost(99999)

	override fun write(buff: FriendlyByteBuf) {
		state.write(buff)
		buff.writeInt(amount)
	}

	override fun play(context: Supplier<NetworkEvent.Context>) {
		context.packetHandled = true
		context.enqueueWork {
			val sender = context.sender ?: return@enqueueWork
			val menu = sender.containerMenu as? MatterPanelMenu ?: return@enqueueWork

			menu.requestReplication(sender, state, amount)
		}
	}
}

class MatterPanelMenu @JvmOverloads constructor(
	p_38852_: Int,
	inventory: Inventory,
	tile: MatterPanelBlockEntity? = null
) : MatteryMenu(MMenus.MATTER_PANEL, p_38852_, inventory, tile), IMatterGraphListener {
	fun taskUpdated(task: MatterTask) {
		sendNetwork(TasksChangePacket(true, listOf(task)))
	}

	fun taskRemoved(task: MatterTask) {
		sendNetwork(TasksChangePacket(false, listOf(task)))
	}

	// client code
	val patterns = ArrayList<PatternState>()
	val tasks = ArrayList<MatterTask>()
	var changeset = 0

	fun networkPatternsUpdated(patterns: Collection<PatternState>) {
		changeset++

		for (pattern in patterns) {
			val index = this.patterns.indexOf(pattern)

			if (index != -1) {
				this.patterns[index] = pattern
			} else {
				this.patterns.add(pattern)
			}
		}
	}

	fun networkPatternsRemoved(patterns: Collection<PatternState>) {
		changeset++

		for (pattern in patterns) {
			this.patterns.remove(pattern)
		}
	}

	fun networkTasksUpdated(tasks: Collection<MatterTask>) {
		changeset++

		for (task in tasks) {
			val index = this.tasks.indexOf(task)

			if (index != -1) {
				this.tasks[index] = task
			} else {
				this.tasks.add(task)
			}

			watcher_update?.accept(task)
		}
	}

	fun networkTasksRemoved(tasks: Collection<MatterTask>) {
		changeset++

		for (task in tasks) {
			this.tasks.remove(task)
			watcher_delete?.accept(task)
		}
	}

	private var watcher_delete: Consumer<MatterTask>? = null
	private var watcher_update: Consumer<MatterTask>? = null

	fun networkTaskWatcher(watcher_update: Consumer<MatterTask>, watcher_delete: Consumer<MatterTask>) {
		this.watcher_delete = watcher_delete
		this.watcher_update = watcher_update
	}

	// server code
	fun requestReplication(ply: ServerPlayer, state: PatternState, how_much: Int) {
		val tile = tile as MatterPanelBlockEntity? ?: return

		val graph = tile.matterGraph ?: return

		if (graph.getPattern(state) == null) {
			OverdriveThatMatters.LOGGER.error("Received replication request from {} of {}, but it is no longer in grid", ply, state)
			return
		}

		tile.addTask(state, how_much)
	}

	fun receiveTaskCancel(ply: ServerPlayer, id: UUID) {
		(tile as MatterPanelBlockEntity?)?.removeTask(id)
	}

	fun requestTaskCancel(id: UUID) {
		MenuNetworkChannel.sendToServer(CancelTaskPacket(id))
	}

	override fun onPatternAdded(state: PatternState) {
		sendNetwork(PatternsChangePacket(true, listOf(state)))
	}

	override fun onPatternRemoved(state: PatternState) {
		sendNetwork(PatternsChangePacket(false, listOf(state)))
	}

	override fun onPatternUpdated(new_state: PatternState, old_state: PatternState) {
		sendNetwork(PatternsChangePacket(true, listOf(new_state)))
	}

	private var initial_send = false

	private var listeningGrid: MatterNetworkGraph? = null

	init {
		if (tile != null) {
			listeningGrid = tile.matterGraph
			tile.attachMenu(this)
			listeningGrid?.addListener(this)
		}
	}

	override fun removed(p_38940_: Player) {
		super.removed(p_38940_)
		(tile as MatterPanelBlockEntity?)?.deatachMenu(this)
		listeningGrid?.removeListener(this)
	}

	override fun broadcastChanges() {
		super.broadcastChanges()
		if (!initial_send) fullPatternBroadcast()
	}

	private fun sendNetwork(packet: Any) {
		MenuNetworkChannel.send(PacketDistributor.PLAYER.with { inventory.player as ServerPlayer }, packet)
	}

	fun fullPatternBroadcast() {
		if (inventory.player.level.isClientSide) {
			initial_send = true
			return
		}

		val tile = tile as MatterPanelBlockEntity?

		if (tile != null) {
			val grid = tile.matterGraph

			if (grid != null) {
				initial_send = true
				sendNetwork(PatternsChangePacket(true, grid.getStoredPatterns()))
			}

			sendNetwork(TasksChangePacket(true, tile.allTasks))
		}
	}

	override val storageSlots: Collection<Slot>
		get() = emptyList()
}
