package ru.dbotthepony.mc.otm.saveddata

import it.unimi.dsi.fastutil.ints.Int2ObjectAVLTreeMap
import it.unimi.dsi.fastutil.ints.Int2ObjectFunction
import net.minecraft.nbt.CompoundTag
import net.minecraft.nbt.IntTag
import net.minecraft.nbt.ListTag
import net.minecraft.nbt.Tag
import net.minecraft.world.level.saveddata.SavedData
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.mc.otm.set

class SavedMapDelegate<V>(val parent: SavedCountingMap<SavedMapDelegate<V>>?, val index: Int, value: V) {
	constructor(value: V) : this(null, -1, value)

	var value: V = value
		set(value) {
			if (field != value) {
				field = value
				parent?.isDirty = true
			}
		}

	companion object {
		fun <V> makeMap(
			serializer: (value: V) -> Tag,
			deserializer: (nbt: Tag) -> V,
			defaultValue: V
		): SavedCountingMap<SavedMapDelegate<V>> {
			return SavedCountingMap(
				serializer = { _, value, _ -> serializer.invoke(value.value) },
				deserializer = { map, nbt, index -> SavedMapDelegate(map, index, deserializer.invoke(nbt)) },
				factory = { map, index -> SavedMapDelegate(map, index, defaultValue) }
			)
		}
	}
}

class SavedCountingMap<T>(
	val serializer: (map: SavedCountingMap<T>, value: T, index: Int) -> Tag,
	val deserializer: (map: SavedCountingMap<T>, nbt: Tag, index: Int) -> T,
	val factory: (map: SavedCountingMap<T>, index: Int) -> T
) : SavedData(), MutableMap<Int, T> {
	var nextIndex = 0
		private set

	fun computeIfAbsent(index: Int): T {
		return map.computeIfAbsent(index, Int2ObjectFunction {
			isDirty = true
			return@Int2ObjectFunction factory.invoke(this, it)
		})
	}

	fun factorize(): T {
		return computeIfAbsent(nextIndex++)
	}

	private val map = Int2ObjectAVLTreeMap<T>()

	override val size: Int
		get() = map.size

	override fun containsKey(key: Int): Boolean {
		return map.containsKey(key)
	}

	override fun containsValue(value: T): Boolean {
		return map.containsValue(value)
	}

	override fun get(key: Int): T? {
		return map[key]
	}

	override fun isEmpty(): Boolean {
		return map.isEmpty()
	}

	override val entries: MutableSet<MutableMap.MutableEntry<Int, T>> by lazy {
		object : MutableSet<MutableMap.MutableEntry<Int, T>> {
			override fun add(element: MutableMap.MutableEntry<Int, T>): Boolean {
				this@SavedCountingMap[element.key] = element.value
				return true
			}

			override fun addAll(elements: Collection<MutableMap.MutableEntry<Int, T>>): Boolean {
				for (value in elements) {
					add(value)
				}

				return true
			}

			override fun clear() {
				this@SavedCountingMap.clear()
			}

			private val setParent = this@SavedCountingMap.map.int2ObjectEntrySet()

			override fun iterator(): MutableIterator<MutableMap.MutableEntry<Int, T>> {
				return object : MutableIterator<MutableMap.MutableEntry<Int, T>> {
					private val parent = setParent.iterator()

					override fun hasNext(): Boolean {
						return parent.hasNext()
					}

					override fun next(): MutableMap.MutableEntry<Int, T> {
						return parent.next()
					}

					override fun remove() {
						parent.remove()
						isDirty = true
					}
				}
			}

			override fun remove(element: MutableMap.MutableEntry<Int, T>): Boolean {
				if (setParent.remove(element)) {
					isDirty = true
					return true
				}

				return false
			}

			override fun removeAll(elements: Collection<MutableMap.MutableEntry<Int, T>>): Boolean {
				if (setParent.removeAll(elements)) {
					isDirty = true
					return true
				}

				return false
			}

			override fun retainAll(elements: Collection<MutableMap.MutableEntry<Int, T>>): Boolean {
				if (setParent.retainAll(elements)) {
					isDirty = true
					return true
				}

				return false
			}

			override val size: Int
				get() = setParent.size

			override fun contains(element: MutableMap.MutableEntry<Int, T>): Boolean {
				return setParent.contains(element)
			}

			override fun containsAll(elements: Collection<MutableMap.MutableEntry<Int, T>>): Boolean {
				return setParent.containsAll(elements)
			}

			override fun isEmpty(): Boolean {
				return setParent.isEmpty()
			}
		}
	}

	override val keys: MutableSet<Int> by lazy {
		object : MutableSet<Int> {
			val parent = this@SavedCountingMap.map.keys

			override fun add(element: Int): Boolean {
				throw UnsupportedOperationException()
			}

			override fun addAll(elements: Collection<Int>): Boolean {
				throw UnsupportedOperationException()
			}

			override fun clear() {
				this@SavedCountingMap.clear()
			}

			override fun iterator(): MutableIterator<Int> {
				return object : MutableIterator<Int> {
					private val parentIterator = parent.iterator()

					override fun hasNext(): Boolean {
						return parentIterator.hasNext()
					}

					override fun next(): Int {
						return parentIterator.nextInt()
					}

					override fun remove() {
						parentIterator.remove()
						isDirty = true
					}
				}
			}

			override fun remove(element: Int): Boolean {
				return this@SavedCountingMap.remove(element) != null
			}

			override fun removeAll(elements: Collection<Int>): Boolean {
				var changes = false

				for (element in elements) {
					changes = remove(element) || changes
				}

				return changes
			}

			override fun retainAll(elements: Collection<Int>): Boolean {
				val iterator = this.iterator()
				var changes = false

				for (element in iterator) {
					if (element !in elements) {
						iterator.remove()
						changes = true
					}
				}

				return changes
			}

			override val size: Int
				get() = parent.size

			override fun contains(element: Int): Boolean {
				return parent.contains(element)
			}

			override fun containsAll(elements: Collection<Int>): Boolean {
				return parent.containsAll(elements)
			}

			override fun isEmpty(): Boolean {
				return parent.isEmpty()
			}
		}
	}

	override val values: MutableCollection<T> by lazy {
		object : MutableCollection<T> {
			private val parent = this@SavedCountingMap.map.values

			override val size: Int
				get() = parent.size

			override fun contains(element: T): Boolean {
				return parent.contains(element)
			}

			override fun containsAll(elements: Collection<T>): Boolean {
				return parent.containsAll(elements)
			}

			override fun isEmpty(): Boolean {
				return parent.isEmpty()
			}

			override fun add(element: T): Boolean {
				throw UnsupportedOperationException()
			}

			override fun addAll(elements: Collection<T>): Boolean {
				throw UnsupportedOperationException()
			}

			override fun clear() {
				this@SavedCountingMap.clear()
			}

			override fun iterator(): MutableIterator<T> {
				return object : MutableIterator<T> {
					private val parentIterator = parent.iterator()

					override fun hasNext(): Boolean {
						return parentIterator.hasNext()
					}

					override fun next(): T {
						return parentIterator.next()
					}

					override fun remove() {
						parentIterator.remove()
						isDirty = true
					}
				}
			}

			override fun remove(element: T): Boolean {
				val indexOf = this@SavedCountingMap.map.firstNotNullOfOrNull { if (it.value == element) it.key else null } ?: return false
				this@SavedCountingMap.remove(indexOf)
				return true
			}

			override fun removeAll(elements: Collection<T>): Boolean {
				var changes = false

				for (element in elements) {
					changes = remove(element) || changes
				}

				return changes
			}

			override fun retainAll(elements: Collection<T>): Boolean {
				val iterator = this.iterator()
				var changes = false

				for (element in iterator) {
					if (element !in elements) {
						iterator.remove()
						changes = true
					}
				}

				return changes
			}
		}
	}

	override fun clear() {
		isDirty = true
		return map.clear()
	}

	override fun put(key: Int, value: T): T? {
		val existing = map.put(key, value)

		if (existing == value) {
			return existing
		}

		isDirty = true
		return existing
	}

	override fun putAll(from: Map<out Int, T>) {
		isDirty = true
		return map.putAll(from)
	}

	override fun remove(key: Int): T? {
		val removed = map.remove(key)

		if (removed != null) {
			isDirty = true
		}

		return removed
	}

	override fun save(output: CompoundTag): CompoundTag {
		output["map"] = ListTag().also {
			for ((key, value) in this.map) {
				val compound = CompoundTag()

				try {
					compound["value"] = this.serializer.invoke(this, value, key)
					it.add(compound)
				} catch(err: Exception) {
					LOGGER.error("Unable to serialize $value at $key", err)
					continue
				}

				compound["key"] = key
			}
		}

		output["next_index"] = nextIndex

		return output
	}

	fun load(input: CompoundTag): SavedCountingMap<T> {
		this.map.clear()

		val map = input["map"] as? ListTag ?: ListTag()
		var maximal = -1

		for (value in map) {
			if (value !is CompoundTag) {
				continue
			}

			val key = (value["key"] as IntTag?)?.asInt ?: continue
			val nbt = value["value"] ?: continue

			try {
				val deserialized = this.deserializer.invoke(this, nbt, key)
				this.map.put(key, deserialized)
			} catch(err: Exception) {
				LOGGER.error("Unable to deserialize value in $this at $key", err)
			}

			maximal = maximal.coerceAtLeast(key)
		}

		nextIndex = (input["next_index"] as IntTag?)?.asInt ?: (maximal + 1)

		isDirty = false
		return this
	}

	companion object {
		private val LOGGER = LogManager.getLogger()
	}
}
